/*
 * Mount handle
 *
 * Copyright (c) 2009-2014, Joachim Metz <joachim.metz@gmail.com>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include "mount_handle.h"
#include "regftools_libcerror.h"
#include "regftools_libclocale.h"
#include "regftools_libcstring.h"
#include "regftools_libregf.h"

/* Creates a mount handle
 * Make sure the value mount_handle is referencing, is set to NULL
 * Returns 1 if successful or -1 on error
 */
int mount_handle_initialize(
     mount_handle_t **mount_handle,
     libcerror_error_t **error )
{
	static char *function = "mount_handle_initialize";

	if( mount_handle == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid mount handle.",
		 function );

		return( -1 );
	}
	if( *mount_handle != NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid mount handle value already set.",
		 function );

		return( -1 );
	}
	*mount_handle = memory_allocate_structure(
	                 mount_handle_t );

	if( *mount_handle == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create mount handle.",
		 function );

		goto on_error;
	}
	if( memory_set(
	     *mount_handle,
	     0,
	     sizeof( mount_handle_t ) ) == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear mount handle.",
		 function );

		goto on_error;
	}
	if( libregf_file_initialize(
	     &( ( *mount_handle )->input_file ),
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to initialize input file.",
		 function );

		goto on_error;
	}
	( *mount_handle )->ascii_codepage = LIBREGF_CODEPAGE_WINDOWS_1252;

	return( 1 );

on_error:
	if( *mount_handle != NULL )
	{
		memory_free(
		 *mount_handle );

		*mount_handle = NULL;
	}
	return( -1 );
}

/* Frees a mount handle
 * Returns 1 if successful or -1 on error
 */
int mount_handle_free(
     mount_handle_t **mount_handle,
     libcerror_error_t **error )
{
	static char *function = "mount_handle_free";
	int result            = 1;

	if( mount_handle == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid mount handle.",
		 function );

		return( -1 );
	}
	if( *mount_handle != NULL )
	{
		if( libregf_file_free(
		     &( ( *mount_handle )->input_file ),
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free input file.",
			 function );

			result = -1;
		}
		memory_free(
		 *mount_handle );

		*mount_handle = NULL;
	}
	return( result );
}

/* Signals the mount handle to abort
 * Returns 1 if successful or -1 on error
 */
int mount_handle_signal_abort(
     mount_handle_t *mount_handle,
     libcerror_error_t **error )
{
	static char *function = "mount_handle_signal_abort";

	if( mount_handle == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid mount handle.",
		 function );

		return( -1 );
	}
	if( mount_handle->input_file != NULL )
	{
		if( libregf_file_signal_abort(
		     mount_handle->input_file,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
			 "%s: unable to signal input file to abort.",
			 function );

			return( -1 );
		}
	}
	return( 1 );
}

/* Sets the ascii codepage
 * Returns 1 if successful or -1 on error
 */
int mount_handle_set_ascii_codepage(
     mount_handle_t *mount_handle,
     const libcstring_system_character_t *string,
     libcerror_error_t **error )
{
	static char *function  = "mount_handle_set_ascii_codepage";
	size_t string_length   = 0;
	uint32_t feature_flags = 0;
	int result             = 0;

	if( mount_handle == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid mount handle.",
		 function );

		return( -1 );
	}
	feature_flags = LIBCLOCALE_CODEPAGE_FEATURE_FLAG_HAVE_KOI8
	              | LIBCLOCALE_CODEPAGE_FEATURE_FLAG_HAVE_WINDOWS;

	string_length = libcstring_system_string_length(
	                 string );

#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
	result = libclocale_codepage_copy_from_string_wide(
	          &( mount_handle->ascii_codepage ),
	          string,
	          string_length,
	          feature_flags,
	          error );
#else
	result = libclocale_codepage_copy_from_string(
	          &( mount_handle->ascii_codepage ),
	          string,
	          string_length,
	          feature_flags,
	          error );
#endif
	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to determine ASCII codepage.",
		 function );

		return( -1 );
	}
	return( result );
}

/* Opens the input of the mount handle
 * Returns 1 if successful or -1 on error
 */
int mount_handle_open_input(
     mount_handle_t *mount_handle,
     const libcstring_system_character_t *filename,
     libcerror_error_t **error )
{
	static char *function = "mount_handle_open_input";

	if( mount_handle == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid mount handle.",
		 function );

		return( -1 );
	}
	if( libregf_file_set_ascii_codepage(
	     mount_handle->input_file,
	     mount_handle->ascii_codepage,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
		 "%s: unable to set ASCII codepage in input file.",
		 function );

		return( -1 );
	}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
	if( libregf_file_open_wide(
	     mount_handle->input_file,
	     filename,
	     LIBREGF_OPEN_READ,
	     error ) != 1 )
#else
	if( libregf_file_open(
	     mount_handle->input_file,
	     filename,
	     LIBREGF_OPEN_READ,
	     error ) != 1 )
#endif
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_IO,
		 LIBCERROR_IO_ERROR_OPEN_FAILED,
		 "%s: unable to open file.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Closes the mount handle
 * Returns the 0 if succesful or -1 on error
 */
int mount_handle_close(
     mount_handle_t *mount_handle,
     libcerror_error_t **error )
{
	static char *function = "mount_handle_close";

	if( mount_handle == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid mount handle.",
		 function );

		return( -1 );
	}
	if( libregf_file_close(
	     mount_handle->input_file,
	     error ) != 0 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_IO,
		 LIBCERROR_IO_ERROR_CLOSE_FAILED,
		 "%s: unable to close input file.",
		 function );

		return( -1 );
	}
	return( 0 );
}

/* Retrieves the key of a specific path
 * Returns 1 if successful, 0 if no such key or -1 on error
 */
int mount_handle_get_key_by_path(
     mount_handle_t *mount_handle,
     const libcstring_system_character_t *path,
     size_t path_length,
     libcstring_system_character_t path_separator,
     libregf_key_t **key,
     libcerror_error_t **error )
{
	libcstring_system_character_t *regf_path = NULL;
	static char *function                    = "mount_handle_get_key_by_path";
	size_t path_index                        = 0;
	size_t regf_path_index                   = 0;
	size_t regf_path_size                    = 0;
	int result                               = 0;

	if( mount_handle == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid mount handle.",
		 function );

		return( -1 );
	}
	if( path == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid path.",
		 function );

		return( -1 );
	}
	if( path_length == 0 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid path length.",
		 function );

		return( -1 );
	}
	if( path_length > (size_t) SSIZE_MAX )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid path length value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( path[ 0 ] != path_separator )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported path.",
		 function );

		return( -1 );
	}
	regf_path_size = ( path_length * 2 ) + 1;

	if( ( regf_path_size > (size_t) SSIZE_MAX )
	 || ( ( sizeof( libcstring_system_character_t ) * regf_path_size ) > (size_t) SSIZE_MAX ) )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid REGF path size value exceeds maximum.",
		 function );

		goto on_error;
	}
	regf_path = libcstring_system_string_allocate(
	             regf_path_size );

	if( regf_path == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create REGF path.",
		 function );

		goto on_error;
	}
	if( path_separator == (libcstring_system_character_t) '/' )
	{
		path_index      = 0;
		regf_path_index = 0;

		while( path_index < path_length )
		{
			if( path[ path_index ] == (libcstring_system_character_t) '/' )
			{
				/* / is changed to \ */
				regf_path[ regf_path_index++ ] = (libcstring_system_character_t) '\\';

				path_index++;
			}
			else if( path[ path_index ] == (libcstring_system_character_t) '\\' )
			{
				if( ( ( path_index + 1 ) < path_length )
				 && ( path[ path_index + 1 ] == (libcstring_system_character_t) '\\' ) )
				{
					/* \\ is changed to \ */
					regf_path[ regf_path_index++ ] = (libcstring_system_character_t) '\\';

					path_index += 2;
				}
				else if( ( ( path_index + 3 ) < path_length )
				      && ( path[ path_index + 1 ] == (libcstring_system_character_t) 'x' )
				      && ( path[ path_index + 2 ] == (libcstring_system_character_t) '2' )
				      && ( path[ path_index + 3 ] == (libcstring_system_character_t) 'f' ) )
				{
					/* \x2f is changed to / */
					regf_path[ regf_path_index++ ] = (libcstring_system_character_t) '/';

					path_index += 4;
				}
				else
				{
					regf_path[ regf_path_index++ ] = path[ path_index++ ];
				}
			}
			else
			{
				regf_path[ regf_path_index++ ] = path[ path_index++ ];
			}
		}
	}
	else if( path_separator == (libcstring_system_character_t) '\\' )
	{
		if( libcstring_system_string_copy(
		     regf_path,
		     path,
		     path_length ) == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_MEMORY,
			 LIBCERROR_MEMORY_ERROR_COPY_FAILED,
			 "%s: unable to copy filename.",
			 function );

			goto on_error;
		}
		regf_path_index = path_length;
	}
	regf_path[ regf_path_index ] = 0;

#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
	result = libregf_file_get_key_by_utf16_path(
		  mount_handle->input_file,
		  (uint16_t *) regf_path,
		  regf_path_index,
		  key,
		  error );
#else
	result = libregf_file_get_key_by_utf8_path(
		  mount_handle->input_file,
		  (uint8_t *) regf_path,
		  regf_path_index,
		  key,
		  error );
#endif
	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve key.",
		 function );

		goto on_error;
	}
	if( path_separator == (libcstring_system_character_t) '/' )
	{
		memory_free(
		 regf_path );
	}
	return( result );

on_error:
	if( path_separator == (libcstring_system_character_t) '/' )
	{
		if( regf_path != NULL )
		{
			memory_free(
			 regf_path );
		}
	}
	return( -1 );
}

/* Retrieves the value of a specific path
 * Returns 1 if successful, 0 if no such value or -1 on error
 */
int mount_handle_get_value_by_path(
     mount_handle_t *mount_handle,
     const libcstring_system_character_t *path,
     size_t path_length,
     libcstring_system_character_t path_separator,
     libregf_key_t **key,
     libregf_value_t **value,
     libcerror_error_t **error )
{
	libcstring_system_character_t *regf_path  = NULL;
	libcstring_system_character_t *value_name = NULL;
	static char *function                     = "mount_handle_get_value_by_path";
	size_t last_path_seperator_index          = 0;
	size_t last_regf_path_seperator_index     = 0;
	size_t path_index                         = 0;
	size_t regf_path_index                    = 0;
	size_t regf_path_size                     = 0;
	size_t value_name_length                  = 0;
	int result                                = 0;

	if( mount_handle == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid mount handle.",
		 function );

		return( -1 );
	}
	if( path == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid path.",
		 function );

		return( -1 );
	}
	if( path_length == 0 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid path length.",
		 function );

		return( -1 );
	}
	if( path_length > (size_t) SSIZE_MAX )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid path length value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( path[ 0 ] != path_separator )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported path.",
		 function );

		return( -1 );
	}
	regf_path_size = ( path_length * 2 ) + 1;

	if( ( regf_path_size > (size_t) SSIZE_MAX )
	 || ( ( sizeof( libcstring_system_character_t ) * regf_path_size ) > (size_t) SSIZE_MAX ) )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid REGF path size value exceeds maximum.",
		 function );

		goto on_error;
	}
	regf_path = libcstring_system_string_allocate(
	             regf_path_size );

	if( regf_path == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create REGF path.",
		 function );

		goto on_error;
	}
	if( path_separator == (libcstring_system_character_t) '/' )
	{
		path_index      = 0;
		regf_path_index = 0;

		while( path_index < path_length )
		{
			if( path[ path_index ] == (libcstring_system_character_t) '/' )
			{
				/* / is changed to \ */
				regf_path[ regf_path_index ] = (libcstring_system_character_t) '\\';

				last_path_seperator_index      = path_index;
				last_regf_path_seperator_index = regf_path_index;

				path_index++;
				regf_path_index++;
			}
			else if( path[ path_index ] == (libcstring_system_character_t) '\\' )
			{
				if( ( ( path_index + 1 ) < path_length )
				 && ( path[ path_index + 1 ] == (libcstring_system_character_t) '\\' ) )
				{
					/* \\ is changed to \ */
					regf_path[ regf_path_index++ ] = (libcstring_system_character_t) '\\';

					path_index += 2;
				}
				else if( ( ( path_index + 3 ) < path_length )
				      && ( path[ path_index + 1 ] == (libcstring_system_character_t) 'x' )
				      && ( path[ path_index + 2 ] == (libcstring_system_character_t) '2' )
				      && ( path[ path_index + 3 ] == (libcstring_system_character_t) 'f' ) )
				{
					/* \x2f is changed to / */
					regf_path[ regf_path_index++ ] = (libcstring_system_character_t) '/';

					path_index += 4;
				}
				else
				{
					regf_path[ regf_path_index++ ] = path[ path_index++ ];
				}
			}
			else
			{
				regf_path[ regf_path_index++ ] = path[ path_index++ ];
			}
		}
		regf_path[ last_regf_path_seperator_index ] = 0;

		path_index      = last_path_seperator_index + 1;
		regf_path_index = last_regf_path_seperator_index + 1;

		while( path_index < path_length )
		{
			if( path[ path_index ] == (libcstring_system_character_t) '/' )
			{
				/* / is changed to \ */
				regf_path[ regf_path_index++ ] = (libcstring_system_character_t) '\\';
			}
			else if( path[ path_index ] == (libcstring_system_character_t) '\\' )
			{
				if( ( ( path_index + 1 ) < path_length )
				 && ( path[ path_index + 1 ] == (libcstring_system_character_t) '\\' ) )
				{
					/* \\ is changed to \ */
					regf_path[ regf_path_index++ ] = (libcstring_system_character_t) '\\';

					path_index += 2;
				}
				else if( ( ( path_index + 3 ) < path_length )
				      && ( path[ path_index + 1 ] == (libcstring_system_character_t) 'x' )
				      && ( path[ path_index + 2 ] == (libcstring_system_character_t) '2' )
				      && ( path[ path_index + 3 ] == (libcstring_system_character_t) 'f' ) )
				{
					/* \x2f is changed to / */
					regf_path[ regf_path_index++ ] = (libcstring_system_character_t) '/';

					path_index += 4;
				}
				else
				{
					regf_path[ regf_path_index++ ] = path[ path_index++ ];
				}
			}
			else
			{
				regf_path[ regf_path_index++ ] = path[ path_index++ ];
			}
		}
		regf_path[ regf_path_index ] = 0;
	}
	else if( path_separator == (libcstring_system_character_t) '\\' )
	{
		if( libcstring_system_string_copy(
		     regf_path,
		     path,
		     path_length ) == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_MEMORY,
			 LIBCERROR_MEMORY_ERROR_COPY_FAILED,
			 "%s: unable to copy filename.",
			 function );

			goto on_error;
		}
		regf_path[ path_length ] = 0;

		last_regf_path_seperator_index = path_length;

		while( last_regf_path_seperator_index > 0 )
		{
			if( path[ last_regf_path_seperator_index ] == (libcstring_system_character_t) '\\' )
			{
				break;
			}
			last_regf_path_seperator_index--;
		}
		regf_path[ last_regf_path_seperator_index ] = 0;

		regf_path_index = last_regf_path_seperator_index + 1;
	}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
	result = libregf_file_get_key_by_utf16_path(
		  mount_handle->input_file,
		  (uint16_t *) regf_path,
		  last_regf_path_seperator_index,
		  key,
		  error );
#else
	result = libregf_file_get_key_by_utf8_path(
		  mount_handle->input_file,
		  (uint8_t *) regf_path,
		  last_regf_path_seperator_index,
		  key,
		  error );
#endif
	if( result != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve key.",
		 function );

		goto on_error;
	}
	last_regf_path_seperator_index += 1;

	value_name        = &( regf_path[ last_regf_path_seperator_index ] );
	value_name_length = regf_path_index - last_regf_path_seperator_index;

	if( value_name_length == 0 )
	{
		memory_free(
		 regf_path );

		return( 0 );
	}
	else if( value_name_length == 9 )
	{
		if( libcstring_system_string_compare(
		     value_name,
		     _LIBCSTRING_SYSTEM_STRING( "(default)" ),
		     9 ) == 0 )
		{
			value_name        = NULL;
			value_name_length = 0;
		}
	}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
	result = libregf_key_get_value_by_utf16_name(
		  *key,
		  (uint16_t *) value_name,
		  value_name_length,
		  value,
		  error );
#else
	result = libregf_key_get_value_by_utf8_name(
		  *key,
		  (uint8_t *) value_name,
		  value_name_length,
		  value,
		  error );
#endif
	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve value from key.",
		 function );

		goto on_error;
	}
	memory_free(
	 regf_path );

	return( result );

on_error:
	if( key != NULL )
	{
		libregf_key_free(
		 key,
		 NULL );
	}
	if( regf_path != NULL )
	{
		memory_free(
		 regf_path );
	}
	return( -1 );
}

